class InconsistentDataException(Exception):
    """
    Raised when (parts of) a Dataset is inconsistent
    """


class DoesNotExistException(Exception):
    """
    Raised when (parts of) a Dataset does not exist
    """
